package snow;

import java.util.*;
import static snow.SWConstants.*;

public class SWGraphModel2
{
	SWModel model;
  // Vertex object corresponding to each intersection
  Vertex intersectionToVertex[][];
  // Vertex object corresponding to each entrance
  Vertex entranceToVertex[][][];
  // Array of all vertices.
  ArrayList<Vertex> vertexToItem;
  double[][] distances;     // Table of minimum distances between vertices.
  Vertex[][] predecessors;  // Predecessors of each vertex in shortest path.
  double[][][] maxDists;  // Shortest distance of the given block to each of its neighbors.
  
  public SWGraphModel2(SWModel _model)
  {
  	model = _model;
  	maxDists = new double[model.width()][model.height()][DIRS];
  	int w = model.width();
  	int h = model.height();
  	int nIntersections = (w + 1) * (h + 1);
  	int nEntrances = w * h * DIRS;
  	vertexToItem = new ArrayList<Vertex>(nIntersections + nEntrances);
  	createVertexMappings();
  	createGraph();
  	// For debugging
  	// printGraph();
  	computeDistances();
  	computeMaxDists();
  }
  
  void printGraph()
  {
    System.out.println("=====");
    for (Vertex u : vertexToItem)
    {
      boolean print = false;
      StringBuilder sb = new StringBuilder(u.toString() + ":");
      for (Vertex v : u.edges)
      {
        if (v.type != Type.ENTRANCE || u.type != Type.ENTRANCE)
        {
          print = true;
          sb.append(" " + v.toString());
        }
      }
      if (print)
      {
        System.out.println(sb.toString());
      }
    }
  }
  
  /**
   * Create vertices for intersections and entrances and insert them into the
   * proper tables.
   */
  void createVertexMappings()
  {
    // Create vertices for each intersection.
  	intersectionToVertex = new Vertex[model.width() + 1][model.height() + 1];
  	for (int ix = 0; ix <= model.width(); ++ix)
  	{
  		for (int iy = 0; iy <= model.height(); ++iy)
  		{
  			int id = vertexToItem.size();
  			Vertex v = new Vertex(id, Type.INTERSECTION, ix, iy, NIL, model.corners());	
  			intersectionToVertex[ix][iy] = v;
  			vertexToItem.add(v);
  		}
  	}

  	// Create vertices for each entrance.
  	entranceToVertex = new Vertex[model.width()][model.height()][DIRS];
  	for (int bx = 0; bx < model.width(); ++bx)
  	{
  		for (int by = 0; by < model.height(); ++by)
  		{
  			for (int dir = 0; dir < DIRS; ++dir)
  			{
  				int id = vertexToItem.size();
  				Vertex v =
  				  new Vertex(id, Type.ENTRANCE, bx, by, dir, model.corners());
  				entranceToVertex[bx][by][dir] = v;
  				vertexToItem.add(v);
  			}
  		}
  	}
  	
    // Initialize distances and predecessors to the proper size and initial
    // value.
  	int n = vertexToItem.size();
  	distances = new double[n][n];
  	predecessors = new Vertex[n][n];
  }
  
  /**
   * Connect the vertices as appropriate.
   */
  void createGraph()
  {
  	connectIntersectionsToIntersections();
  	
  	if (model.walkThrough())
  	{
    	connectEntrancesToEntrances();  		
  	}
  	
  	if (model.corners())
  	{
  		connectCornerEntrancesToIntersections();
  	}
  	else
  	{
    	connectEdgeEntrancesToIntersections();  		
  	}
  }
  
  /**
   * Connect intersections with each other
   * based on the paths plowed.
   */
  void connectIntersectionsToIntersections()
  {
  	for (int ix = 0; ix <= model.width(); ++ix)
  	{
  		for (int iy = 0; iy <= model.height(); ++iy)
  		{
        if (iy < model.height() && model.ns(ix, iy))
        {
          Vertex north = iToV(ix, iy);
          Vertex south = iToV(ix, iy + 1);
          join(north, south, I_TO_I_COST);
        }
        if (ix < model.width() && model.ew(ix, iy))
        {
          Vertex west = iToV(ix, iy);
          Vertex east = iToV(ix + 1, iy);
          join(east, west, I_TO_I_COST);
        }
  		}
  	}
  }
  
  /**
   * Connect entrances within a block to each other.
   * (Appropriate for when walkthroughs are allowed.
   */
  void connectEntrancesToEntrances()
  {
  	for (int bx = 0; bx < model.width(); ++bx)
  	{
  		for (int by = 0; by < model.height(); ++by)
  		{
  			for (int dir1 = 0; dir1 < DIRS; ++dir1)
  			{
  				Vertex v1 = eToV(bx, by, dir1);
  				for (int dir2 = 0; dir2 < DIRS; ++dir2)
  				{
  					if (dir1 != dir2)
  					{
  						Vertex v2 = eToV(bx, by, dir2);
  						double cost = eToECost(dir1, dir2);
  						join(v1, v2, cost);
  					}
  				}
  			}
  		}
  	}
  }
  
  /**
   * Connect corner entrances to their intersections
   * when those intersections have been plowed.
   */
  void connectCornerEntrancesToIntersections()
  {
  	for (int ix = 0; ix <= model.width(); ++ix)
  	{
  		for (int iy = 0; iy <= model.height(); ++iy)
  		{
  			if (model.intersectionPlowed(ix, iy))
  			{
  			  Vertex iv = iToV(ix, iy);
  			  // Find the four adjacent corner entrances.
  				Vertex[] vs = new Vertex[DIRS];
  				vs[NW] = eToV(ix - 1, iy - 1, SE);
  				vs[NE] = eToV(ix, iy - 1, SW);
  				vs[SE] = eToV(ix, iy, NW);
  				vs[SW] = eToV(ix - 1, iy, NE);
  				for (Vertex v : vs)
  				{
						if (v != null)
						{
							join(v, iv, E_TO_I_COST);
						}
  				}
  			}
  		}
  	}
  }

  /**
   * For streets that have been plowed, connect the two
   * adjacent edge entrances to each other and
   * to the nearest two intersections.
   */
  void connectEdgeEntrancesToIntersections()
  {
  	for (int ix = 0; ix <= model.width(); ++ix)
  	{
  		for (int iy = 0; iy <= model.height(); ++iy)
  		{
  			if (iy < model.height() && model.ns(ix, iy))
  			{
  				Vertex eLeft = eToV(ix - 1, iy, E);
  				Vertex eRight = eToV(ix, iy, W);
  				Vertex iNorth = iToV(ix, iy);
  				Vertex iSouth = iToV(ix, iy + 1);
  				safeJoin(eLeft, eRight, 2 * E_TO_I_COST);
  				safeJoin(eLeft, iNorth, 0.5 * I_TO_I_COST);
  				safeJoin(eLeft, iSouth, 0.5 * I_TO_I_COST);
   				safeJoin(eRight, iNorth, 0.5 * I_TO_I_COST);
  				safeJoin(eRight, iSouth, 0.5 * I_TO_I_COST);
  			}
  			if (ix < model.width() && model.ew(ix, iy))
  			{
  				Vertex eTop = eToV(ix, iy - 1, S);
  				Vertex eBottom = eToV(ix, iy, N);
  				Vertex iWest = iToV(ix, iy);
  				Vertex iEast = iToV(ix + 1, iy);
  				safeJoin(eTop, eBottom, 2 * E_TO_I_COST);
  				safeJoin(eTop, iWest, 0.5 * I_TO_I_COST);
  				safeJoin(eTop, iEast, 0.5 * I_TO_I_COST);
   				safeJoin(eBottom, iWest, 0.5 * I_TO_I_COST);
  				safeJoin(eBottom, iEast, 0.5 * I_TO_I_COST);
  			}
  		}
  	}
  }
  
  /**
   * Return whether to compute distances for this block
   * or rely on its neighbors ("checkerboard optimization").
   * @param bx x-coordinate of the block
   * @param by y-coordinate of the block
   * @return whether to compute distances directly. 
   */
  boolean shouldComputeDistance(int bx, int by)
  {
    return (bx + by) % 2 == 1;
  }
  
  void computeDistances()
  {
  	for (Vertex v : vertexToItem)
  	{
  		if (v.type == Type.ENTRANCE && shouldComputeDistance(v.vx, v.vy))
  		{
  			doDijkstra(v);
  		}
  	}
  }
  
  /**
   * Return the direction in which dest is a neighbor of source
   * or NIL if dest is not adjacent to source.
   * @param source the source vertex (block)
   * @param dest the destination vertex
   * @return the cardinal direction or NIL if not adjacent
   */
  int whichNeighbor(Vertex source, Vertex dest)
  {
    if (dest.type != Type.ENTRANCE)
    {
      return NIL;
    }
    if (source.vx == dest.vx)
    {
      if (source.vy - 1 == dest.vy)
      {
        return N;
      }
      else if (source.vy + 1 == dest.vy)
      {
        return S;
      }
    }
    else if (source.vy == dest.vy)
    {
      if (source.vx - 1 == dest.vx)
      {
        return W;
      }
      else if (source.vx + 1 == dest.vx)
      {
        return E; 
      }
    }
    return NIL;
  }
  
  static boolean allTrue(boolean[] bb)
  {
    for (boolean b : bb)
    {
      if (!b)
      {
        return false;
      }
    }
    return true;
  }
  
  // See CLRS p. 595
  // Terminate early if:
  // 1. vertex pulled off Q has infinite distance (means graph not connected)
  // 2. all the adjacent blocks to the current block have been accounted for.
  void doDijkstra(Vertex s)
  {
  	// INITIALIZE-SINGLE-SOURCE(G, s)
  	double d[] = distances[s.id];
  	Vertex[] pred = predecessors[s.id];
  	Arrays.fill(d, Double.POSITIVE_INFINITY);
  	d[s.id] = 0;
  	
  	// Initialize S and Q
  	// Set<Vertex> S = new HashSet<Vertex>(); NOT NEEDED
  	MinQueue Q = new MinQueue(vertexToItem, s.id);

  	// whether blocks in each dir are done
  	boolean[] neighbors = new boolean[DIRS]; 
  	while (!Q.isEmpty())
  	{
  		Vertex u = Q.poll();
  		if (d[u.id] == Double.POSITIVE_INFINITY)
  		{
  		  return;
  		}
  		int which = whichNeighbor(s, u);
  		if (which != NIL)
  		{
  		  neighbors[which] = true;
  		}
  		if (allTrue(neighbors))
  		{
  		  return;
  		}
  		
  		// S.add(u); NOT NEEDED
  		
  		// For each vertex v in Adj[u]
  		for (int i = 0; i < u.edges.size(); ++i)
  		{
  			// RELAX(u, v, w)
  			Vertex v = u.edges.get(i);
  			double tmp = d[u.id] + u.weights.get(i);
  			if (d[v.id] > tmp)
  			{
  				d[v.id] = tmp;
  				pred[v.id] = u;
  			}
  		}
  	}
  }
  
  /**
   * Return the Block adjacent to the given block in the given direction.
   * @param b the start block
   * @param dir the direction to move
   * @return the adjacent block or null if no block in that direction
   */
  Block getAdjacentBlock(Block b, int dir)
  {
    int bx = b.bx;
    int by = b.by;
    switch (dir)
    {
      case N:
        return by == 0 ? null : new Block(bx, by - 1);
      case E:
        return bx == (model.width() - 1) ? null : new Block(bx + 1, by);
      case S:
        return by == (model.height() - 1) ? null : new Block(bx, by + 1);
      case W:
        return bx == 0 ? null : new Block(bx - 1, by);
      default:
        return null;
    }
  }
  
  /**
   * Get the shortest path from the block in the given direction
   * to the specified block.
   * @param b The destination block.
   * @param dir The direction of the starting block
   * @return a list of vertices on the route, including
   * the start and end blocks.
   */
  List<Vertex> getRoute(Block b, int dir)
  {
    Block adj = getAdjacentBlock(b, dir);
    if (adj == null)
    {
      // No block in that direction.
      return null;
    }
    return
      shouldComputeDistance(b.bx, b.by) ?
        doGetRoute(b, adj) :
        doGetRoute(adj, b);
  }
  
  List<Vertex> doGetRoute(Block source, Block dest)
  {
    Solution s = minDist(source.bx, source.by, dest.bx, dest.by);
    if (s.dist == Double.POSITIVE_INFINITY)
    {
      // No route from block in that direction.
      return null;
    }
    // Walk the predecessor chain.
    List<Vertex> ret = new ArrayList<Vertex>();
    Vertex[] pred = predecessors[s.src.id];
    Vertex tmp = s.dest;
    do
    {
      ret.add(tmp);
      tmp = pred[tmp.id];
    }
    while (tmp != null);
    return ret;
  }
  
  /**
   * Return the cost of going between the entrances in the
   * two given directions within a block.
   * @param dir1 the first direction
   * @param dir2 the second direction
   * @return the cost (distance) between the two entrances
   */
  double eToECost(int dir1, int dir2)
  {
    if (dir1 == dir2)
    {
      return 0;
    }
    boolean opposite = Math.abs(dir1 - dir2) == 2;
    int tmp = model.corners() ? 0 : 2;
    tmp += opposite ? 1 : 0;
    return model.walkThroughCost(tmp);
  }
  
  /**
   * Lookup the vertex corresponding to the given intersection
   * @param ix the x-coordinate of the intersection
   * @param iy the y-coordinate of the intersection
   * @return the corresponding vertex
   */
  Vertex iToV(int ix, int iy)
  {
  	if (ix >= 0 && ix <= model.width() && iy >= 0 && iy <= model.height())
  	{
    	return intersectionToVertex[ix][iy];  		
  	}
  	return null;
  }
  
  /**
   * Lookup the vertex corresponding to the entrance
   * of the given block in the given direction.
   * @param bx the x-coordinate of the block
   * @param by the y-coordinate of the block
   * @param dir the direction of the entrance
   * @return the corresponding vertex
   */
  Vertex eToV(int bx, int by, int dir)
  {
  	if (bx >= 0 && bx < model.width() && by >= 0 && by < model.height())
  	{
    	return entranceToVertex[bx][by][dir];
  	}
  	return null;
  }
  
  /**
   * Connect the given vertices with an edge of the given cost
   * provided that both vertices are non-null.
   * @param v1 the first vertex
   * @param v2 the second vertex
   * @param cost the cost (weight) of the edge
   */
  void safeJoin(Vertex v1, Vertex v2, double cost)
  {
  	if (v1 != null && v2 != null)
  	{
  		join(v1, v2, cost);
  	}
  }
  
  /**
   * Connect the given vertices with an edge of the given cost.
   * @param v1 the first vertex
   * @param v2 the second vertex
   * @param cost the cost (weight) of the edge
   */
  void join(Vertex v1, Vertex v2, double cost)
  {
  	v1.addEdge(v2, cost);
  	v2.addEdge(v1, cost);
  }
  
  /**
   * Return the maximum distance of the given block to each
   * of its adjacent blocks.
   * @param bx the x-coordinate of the block
   * @param by the y-coordinate of the block
   * @return an array of the maximum distances in each direction
   */
	public double[] maxDists(int bx, int by)
	{
	  return maxDists[bx][by];
	}
	
	void computeMaxDists()
	{
	  for (int bx = 0; bx < model.width(); ++bx)
	  {
	    for (int by = 0; by < model.height(); ++by)
	    {
	      Block b = new Block(bx, by);
	      for (int dir = 0; dir < DIRS; ++dir)
	      {
	        double dist = Double.POSITIVE_INFINITY;
	        Block adj = getAdjacentBlock(b, dir);
	        if (adj == null)
	        {
	          dist = -1;  // Special val to indicate no block.
	        }
	        else
	        {
	          Solution s = minDist(b.bx, b.by, adj.bx, adj.by);
	          if (s != null)
	          {
	            dist = s.dist;
	          }
	        }
          maxDists[bx][by][dir] = dist;
	      }
	    }
	  }
	}
	
	/**
	 * Return the length of the shortest path between
	 * the two specified blocks.
	 * @param bx1 the x-coordinate of the first block
	 * @param by1 the y-coordinate of the first block
	 * @param bx2 the x-coordinate of the second block
	 * @param by2 the y-coordinate of the second block
	 * @return the length of the shortest path or
	 * infinity if there is no path
	 */
	Solution minDist(int bx1, int by1, int bx2, int by2)
	{
	  if (shouldComputeDistance(bx1, by1))
	  {
	    return doMinDist(bx1, by1, bx2, by2);
	  }
	  else if (shouldComputeDistance(bx2, by2))
	  {
	    return doMinDist(bx2, by2, bx1, by1);
	  }
	  throw new RuntimeException("blocks are not adjacent");
	}
		
	Solution doMinDist(int bx1, int by1, int bx2, int by2)
	{
		Solution ret = new Solution();
		ArrayList<Vertex> vs1 = new ArrayList<Vertex>();
		ArrayList<Vertex> vs2 = new ArrayList<Vertex>();
		for (int dir = 0; dir < DIRS; ++dir)
		{
			vs1.add(eToV(bx1, by1, dir));
			vs2.add(eToV(bx2, by2, dir));
		}
		for (Vertex v1 : vs1)
		{
			for (Vertex v2 : vs2)
			{
				double dist = distances[v1.id][v2.id];
				if (dist < ret.dist)
				{
					ret.dist = dist;
					ret.src = v1;
					ret.dest = v2;
				}					
			}
		}
		return ret;
	}
	
	/**
	 * A struct representing a solution to a shortest path problem.
	 * Contains the start and end vertices and the distance.
	 */
	static class Solution
	{
	  Vertex src;
	  Vertex dest;
	  double dist;
	  
	  Solution()
	  {
	    src = null;
	    dest = null;
	    dist = Double.POSITIVE_INFINITY;
	  }
	}
	
	class MinQueue
	{
	  LinkedList<Vertex> q = new LinkedList<Vertex>();
	  int index;
	  
	  MinQueue(Collection<Vertex> start, int _index)
	  {
	    index = _index;
	    q.addAll(start);
	  }
	  
	  boolean isEmpty()
	  {
	    return q.isEmpty();
	  }
	  
	  // Linear scan
	  // Binary min-heap probably not worth it for this size.
	  Vertex poll()
	  {
	    if (isEmpty())
	    {
	      return null;
	    }
	    Vertex ret = q.getFirst();
	    double retdist = distances[index][ret.id];
	    for (Vertex v : q)
	    {
	      double vdist = distances[index][v.id];
	      if (vdist < retdist)
	      {
	        ret = v;
	        retdist = vdist;
	      }
	    }
	    q.remove(ret);
	    return ret;
	  }
	}
}
