package snow;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import static snow.SWConstants.*;

public class SWMainPanel
  extends JPanel
  implements ChangeListener, FocusListener
{
  static final int SIZE = 50;    // size in pixels of each block
  static final int STREETWIDTH = 6;  // width in pixels of each street
  // inner size of each block
  static final int INNERSIZE = SIZE - STREETWIDTH - 1;
  static final int ESIZE = 2;  // size of entrances
  // offset of entrances from north and west sides
  static final int ENWOFF = 3;
  // offset of entrances from east and south sides
  static final int ESEOFF = INNERSIZE - ENWOFF - ESIZE;
  // offset of entrances in the middle
  static final int EMIDOFF = (INNERSIZE - ESIZE) / 2;
  static final int XCO[] = { ENWOFF, ESEOFF, ESEOFF, ENWOFF };
  static final int XMO[] = { EMIDOFF, ESEOFF, EMIDOFF, ENWOFF };
  static final int PLOWWIDTH = STREETWIDTH + 1;
  static final int PATHWIDTH = STREETWIDTH - 2;
  static final int FOCUSBORDER = 5;
  static final Color LIGHTGRAY = new Color(220, 220, 220);
  static final Color ROUTECOLOR = Color.RED;
  static final Color INDCOLOR = Color.GRAY;
  
  // Gap around indicators (when entrances on sides)
  static final int IND_GAP = ENWOFF;
  // How much to adjust for height of indicators
  static final int IND_HA = 2;
  // How much to adjust for width of indicators
  static final int IND_WA = 5;
  // Offset for indicator in center
  static final int IND_C_OFF = (INNERSIZE - IND_WA) / 2;
  // Offset for indicator on near side
  static final int IND_N_OFF = IND_GAP + IND_HA;
  // Offset for indicator on far side
  static final int IND_F_OFF = INNERSIZE - IND_GAP - IND_HA;
  // Offset of first set of indicators when entrances are on sides
  static final int IND_S1_OFF = ENWOFF + ESIZE + IND_GAP;
  // Offset of second set of indicators when entrances are on sides
  static final int IND_S2_OFF = INNERSIZE - IND_S1_OFF - IND_WA;

  // x offsets for indicators when entrances are on corners
  static final int XIOC[] = { IND_C_OFF, IND_F_OFF, IND_C_OFF, IND_N_OFF, };
  // x offsets for first set of indicators when entrances are on sides
  static final int XIOS1[] = { IND_S1_OFF, IND_F_OFF, IND_S1_OFF, IND_N_OFF, };
  // x offsets for second set of indicators when entrances are on sides
  static final int XIOS2[] = { IND_S2_OFF, IND_F_OFF, IND_S2_OFF, IND_N_OFF, };
  
  SWGUIModel guiModel;
  
  int charHeightFudge;
  int charWidthFudge;
  	
	public SWMainPanel(SWGUIModel _guiModel)
	{
		guiModel = _guiModel;
		guiModel.addChangeListener(this);
		
		setOpaque(true);
		setBackground(Color.WHITE);
		setBorder(BorderFactory.createLineBorder(Color.black));
		setFocusable(true);
		addMouseListener(new SWMainMouseListener());
		addFocusListener(this);
		initActions();
	}
	
	void initActions()
	{
		InputMap im = getInputMap();
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "plowLeft");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "plowRight");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "plowUp");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "plowDown");
		im.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "togglePlow");
		ActionMap am = getActionMap();
		am.put("plowLeft", new LeftAction());
		am.put("plowRight", new RightAction());
		am.put("plowUp", new UpAction());
		am.put("plowDown", new DownAction());
		am.put("togglePlow", new TogglePlowAction());
	}
	
  public Dimension getPreferredSize()
  {
    // + 2 in each dimension for the border 
    return
      new Dimension(
        SIZE * (guiModel.width() + 2),
        SIZE * (guiModel.height() + 2));
  }
  
  /**
   * Get the offset in pixels corresponding to the given intersection
   * coordinate.
   * @param i the coordinate of the intersection
   * @return the offset from the top or left in pixels
   */
  int getIntersectionOffset(int i)
  {
    return (i + 1) * SIZE;
  }

  /**
   * Return the geometric point corresponding to the center of the given
   * intersection.
   * @param ix x-coordinate of the intersection
   * @param iy y-coordinate of the intersection
   * @return the corresponding pixel location
   */
  Point getIntersectionPoint(int ix, int iy)
  {
    return new Point(getIntersectionOffset(ix), getIntersectionOffset(iy));
  }
  
  /**
   * Return the geometric point of the upper-left (NW) corner of the
   * given block.
   * @param bx x-coordinate of the block
   * @param by y-coordinate of the block
   * @return the corresponding pixel location
   */
  Point getBlockNWCornerPoint(int bx, int by)
  {
    Point p = getIntersectionPoint(bx, by);
    return new Point(p.x + STREETWIDTH/2, p.y + STREETWIDTH/2);
  }
  
  /**
   * Get the geometric rectangle occupied by the given block
   * @param bx x-coordinate of the block
   * @param by y-coordinate of the block
   * @return the corresponding rectangle
   */
  Rectangle getBlockRectangle(int bx, int by)
  {
    Point p = getBlockNWCornerPoint(bx, by);
    return new Rectangle(p.x, p.y, INNERSIZE, INNERSIZE);
  }
  
  
  /**
   * Get the point corresponding to the (upper-left corner of the)
   * entrance to the given block in the given direction.
   * @param bx x-coordinate of the block
   * @param by y-coordinate of the block
   * @param dir direction of the entrance
   * @return the corresponding pixel location
   */
  Point getEntrancePoint(int bx, int by, int dir)
  {
    Point p = getBlockNWCornerPoint(bx, by);
    int xo[] = guiModel.corners() ? XCO : XMO;  // x offsets for entrances
    int yindex = (dir + 3) % xo.length;  // y offset is 3 over from x offset
    return new Point(p.x + xo[dir], p.y + xo[yindex]);
  }
  
  /**
   * Get the point of the upper-left corner of the text box
   * for the given block.
   * @param bx x-coordinate of the block
   * @param by y-coordinate of the block
   * @return the corresponding pixel location.
   */
  Point getTextPoint(int bx, int by, String s)
  {
    int n = s.length();
    if (s.contains("."))
    {
      --n;
    }
    Point p = getBlockNWCornerPoint(bx, by);
    return
      new Point(
        p.x + INNERSIZE/2 - n*charWidthFudge,
        p.y + INNERSIZE/2 + charHeightFudge);    
  }
  
  public void paintComponent(Graphics g)
  {
  	predraw(g);
    drawLabels(g);
    drawBlocks(g);
    drawPaths(g);
    drawPlow(g);
    drawRoutes(g);
    drawInfo(g);
  }

  /**
   * Initialization, background, and border for painting the grid.
   * @param g the graphics context
   */
  void predraw(Graphics g)
  {
    FontMetrics fm = g.getFontMetrics();
    charHeightFudge = fm.getHeight()/2;
    charWidthFudge = fm.charWidth('0')/2;

  	g.setColor(getBackground());
  	g.fillRect(0, 0, getWidth(), getHeight());
  	if (isFocusOwner())
  	{
  		g.setColor(Color.GRAY);
  		Dimension dim = getPreferredSize();
  		g.drawRect(
        FOCUSBORDER,
        FOCUSBORDER,
        dim.width - 2*FOCUSBORDER - 1,
        dim.height - 2*FOCUSBORDER - 1);
  	}
  	g.setColor(getForeground());  	
  }
  
  /**
   * Draw the zero-based numberings of the blocks along the top and left.
   * @param g the graphics context
   */
  void drawLabels(Graphics g)
  {
    for (int bx = 0; bx < guiModel.width(); ++bx)
    {
      int x = getIntersectionOffset(bx) + SIZE/2 - charWidthFudge;
      g.drawString(Integer.toString(bx), x, SIZE/2);
    }
    
    for (int by = 0; by < guiModel.height(); ++by)
    {
      int y = getIntersectionOffset(by) + SIZE/2 + charHeightFudge;
      g.drawString(Integer.toString(by), SIZE/3, y);
    }
  }
  
  /**
   * Draw the blocks themselves, including entrances and distances.
   * @param g the graphics context
   */
  void drawBlocks(Graphics g)
  {
    for (int bx = 0; bx < guiModel.width(); ++bx)
    {
      for (int by = 0; by < guiModel.height(); ++by)
      {
        BlockRenderer br = new BlockRenderer(g, bx, by);
        br.render();
      }
    }
  }

  /**
   * Return the maximum value in the array
   * (ignoring negative values).
   * @param vals the array
   * @return the maximum, or 0 if all values are negative
   */
  static double max(double[] vals)
  {
    double ret = 0;
    for (double d : vals)
    {
      if (d > ret)
      {
        ret = d;
      }
    }
    return ret;
  }
  
  /**
   * Draw the location and state of the plow.
   * @param g the graphics context
   */
  void drawPlow(Graphics g)
  { 
    int x = getIntersectionOffset(guiModel.plowX()) - PLOWWIDTH/2 - 1;
    int y = getIntersectionOffset(guiModel.plowY()) - PLOWWIDTH/2 - 1;
    Color old = g.getColor();
    g.setColor(guiModel.plowing() ? Color.RED : Color.GREEN);
    g.fillRect(x, y, PLOWWIDTH, PLOWWIDTH);
    g.setColor(old);
    g.drawRect(x, y, PLOWWIDTH, PLOWWIDTH);
  }
  
  /**
   * Draw the currently plowed path(s).
   * @param g the graphics context
   */
  void drawPaths(Graphics g)
  {
    Color c = g.getColor();
    if (guiModel.isBlockSelected())
    {
      // Lightly shade the path if we are in "inspect" mode.
      g.setColor(Color.GRAY); 
    }
    
    // Draw all north-south plowed streets.
    for (int ix = 0; ix <= guiModel.width(); ++ix)
    {
      for (int iy = 0; iy < guiModel.height(); ++iy)
      {
        if (guiModel.ns(ix, iy))
        {
          int x = getIntersectionOffset(ix) - PATHWIDTH/2;
          int y = getIntersectionOffset(iy) - PATHWIDTH/2;
          g.fillRect(x, y, PATHWIDTH, SIZE + PATHWIDTH);
        }
      }
    }
    
    // Draw all east-west plowed streets.
    for (int ix = 0; ix < guiModel.width(); ++ix)
    {
      for (int iy = 0; iy <= guiModel.height(); ++iy)
      {
        if (guiModel.ew(ix, iy))
        {
          int x = getIntersectionOffset(ix) - PATHWIDTH/2;
          int y = getIntersectionOffset(iy) - PATHWIDTH/2;
          g.fillRect(x, y, SIZE+PATHWIDTH, PATHWIDTH);
        }
      }
    }
    g.setColor(c);
  }
  
  /**
   * Draw shortest paths if we are in inspect mode.
   * @param g the graphics context
   */
  void drawRoutes(Graphics g)
  {
    for (int dir = 0; dir < DIRS; ++dir)
    {
      drawRoute(g, dir);
    }
  }
  
  /**
   * Draw the shortest path to the selected block from the adjacent
   * block in the given direction.
   * @param g the graphics context
   * @param dir the direction to the adjacent block
   */
  void drawRoute(Graphics g, int dir)
  {
    java.util.List<Vertex> route = guiModel.getSelectedRoute(dir);
    if (route == null)
    {
      // No route or no selected block.
      return;
    }
    int n = route.size();
    int[] xPoints = new int[n];
    int[] yPoints = new int[n];
    for (int i = 0; i < n; ++i)
    {
      Vertex v = route.get(i);
      // Find the pixel location of the vertex and add it to the list.
      Point p =
        v.type == Type.ENTRANCE ?
            getEntrancePoint(v.vx, v.vy, v.dir) :
            getIntersectionPoint(v.vx, v.vy);
      xPoints[i] = p.x;
      yPoints[i] = p.y;
    }
    Color c = g.getColor();
    g.setColor(ROUTECOLOR);
    g.drawPolyline(xPoints, yPoints, n);
    g.setColor(c);
  }

  /**
   * Draw the current status message if any.
   * @param g the graphics context
   */
  void drawInfo(Graphics g)
  {
  	String msg = guiModel.message();
  	if (msg != null)
  	{
      int x = getIntersectionOffset(0);
      int y = getIntersectionOffset(guiModel.height()) + SIZE/2;
      Color old = g.getColor();
      g.setColor(Color.RED);
      g.drawString(msg, x, y);
      g.setColor(old);
  	}
  }

  private class BlockRenderer
  {
    Graphics g;
    int bx;
    int by;
    Rectangle rect;
    Point p;
    double[] mds;
    double md;
    
    BlockRenderer(Graphics _g, int _bx, int _by)
    {
      g = _g;
      bx = _bx;
      by = _by;
      rect = getBlockRectangle(bx, by);
      p = getBlockNWCornerPoint(bx, by);
      mds = guiModel.maxDists(bx, by);
      md = max(mds);
    }
    
    void render()
    {
      if (guiModel.isBlockSelected(bx, by))
      {
        shadeRect();
      }
      drawBlockBoundary();
      drawEntrances();
      showMaxDist();
      indicateOffendingBlocks();      
    }
    
    void shadeRect()
    {
      Color c = g.getColor();
      g.setColor(LIGHTGRAY);
      g.fillRect(rect.x, rect.y, rect.width, rect.height);
      g.setColor(c);
    }

    void drawBlockBoundary()
    {
      g.drawRect(rect.x, rect.y, rect.width, rect.height); 
    }
    
    void drawEntrances()
    {
      for (int dir = 0; dir < DIRS; ++dir)
      {
        Point p = getEntrancePoint(bx, by, dir);
        g.drawRect(p.x, p.y, ESIZE, ESIZE);
      }   
    }

    /**
     * Show which block(s) are the cause of the maximum distance.
     */
    void indicateOffendingBlocks()
    {
      if (guiModel.isBlockSelected())
      {
        // Don't show offending blocks in "inspect" mode.
        return;
      }
      if (md == 0)
      {
        // No offending blocks.
        return;
      }
      boolean hasFinite = false;
      for (double d : mds)
      {
        if (d != Double.POSITIVE_INFINITY && d != -1)
        {
          hasFinite = true;
        }
      }
      if (!hasFinite)
      {
        // Don't clutter with inaccessible blocks.
        return;
      }
      Color c = g.getColor();
      g.setColor(INDCOLOR);
      for (int dir = 0; dir < DIRS; ++dir)
      {
        if (mds[dir] == md)
        {
          if (guiModel.corners())
          {
            drawIndicator(p, dir, XIOC);
          }
          else
          {
            drawIndicator(p, dir, XIOS1);
            drawIndicator(p, dir, XIOS2);
          }
        }
      }
      g.setColor(c);
    }
    
    void drawIndicator(Point p, int dir, int[] offsets)
    {
      boolean horizontal = dir == N || dir == S;
      boolean positive = dir == E || dir == S;
      int bias = positive ? 1 : -1;
      int x[] = new int[3];
      int y[] = new int[3];
      int xd[] = new int[3];
      int yd[] = new int[3];

      x[0] = offsets[dir];
      y[0] = offsets[(dir + 3) % DIRS];
      xd[0] = horizontal ? 4 : 0;
      yd[0] = horizontal ? 0 : 4;
      x[1] = horizontal ? x[0] + 1 : x[0] + bias;
      y[1] = horizontal ? y[0] + bias : y[0] + 1;
      xd[1] = horizontal ? 2 : 0;
      yd[1] = horizontal ? 0 : 2;
      x[2] = horizontal ? x[0] + 2 : x[0];
      y[2] = horizontal ? y[0] : y[0] + 2;
      xd[2] = horizontal ? 0 : bias * 2;
      yd[2] = horizontal ? bias * 2 : 0;
      
      for (int i = 0; i < 3; ++i)
      {
        g.drawLine(p.x + x[i], p.y + y[i], p.x + x[i] + xd[i], p.y + y[i] + yd[i]);
      }
    }
    
    void showMaxDist()
    {
      String mdstr = SWGUIModel.formatDouble(md);
      Point tp = getTextPoint(bx, by, mdstr);
      g.drawString(mdstr, tp.x, tp.y);    
    }  
  }

  class UpAction extends AbstractAction
  {
		public void actionPerformed(ActionEvent e)
		{
			guiModel.plowUp();
		}
  }

  class DownAction extends AbstractAction
  {
		public void actionPerformed(ActionEvent e)
		{
			guiModel.plowDown();
		}
  }

  class LeftAction extends AbstractAction
  {
		public void actionPerformed(ActionEvent e)
		{
			guiModel.plowLeft();
		}
  }

  class RightAction extends AbstractAction
  {
		public void actionPerformed(ActionEvent e)
		{
			guiModel.plowRight();
		}
  }

  class TogglePlowAction extends AbstractAction
  {
    public void actionPerformed(ActionEvent e)
    {
    	guiModel.setPlowing(!guiModel.plowing());
    }
  }
  
	public void stateChanged(ChangeEvent e)
	{
		repaint();
	}

	public void focusGained(FocusEvent e)
	{
		repaint();
	}

	public void focusLost(FocusEvent e)
	{
		repaint();
	}
	
	class SWMainMouseListener extends MouseAdapter
	{
		public void mouseEntered(MouseEvent e)
		{
      requestFocusInWindow();
    }
		
		/**
		 * Scan the blocks to see which block (if any)
		 * was selected, and toggle its selection state.
		 */
		public void mouseClicked(MouseEvent e)
		{
		  Point p = e.getPoint();
		  for (int bx = 0; bx < guiModel.width(); ++bx)
		  {
		    for (int by = 0; by < guiModel.height(); ++by)
		    {
		      Rectangle rect = getBlockRectangle(bx, by);
		      if (rect.contains(p))
		      {
		        guiModel.selectBlock(bx, by);
		        return;
		      }
		    }
		  }
		}
	}
}
