package snow;

import static snow.SWConstants.*;

public class SWModel
{
	int width;   // number of blocks going east to west
	int height;  // number of blocks going north to south
	// There are width+1 north-south streets,
  // indexed from the left starting at 0.
	// There are height+1 east-west streets, 
  // indexed from the top starting at 0.
	boolean corners;
	boolean walkThrough;
	boolean ns[][];  // ns[x][y] = true means that the xth north-south street
	                 // is plowed between the yth and (y+1)th east-west street
	boolean ew[][];  // ew[x][y] = true means that the yth east-west street
                   // is plowed between the xth and (x+1)th east-west street
	SWGraphModel2 graph = null;
	double[] costs = new double[COSTS];

	public SWModel
	(
		int _width,
		int _height,
		boolean _corners,
		boolean _walkThrough
	)
	{
		width = _width;
		height = _height;
		corners = _corners;
		walkThrough = _walkThrough;
		ns = new boolean[width + 1][height];
		ew = new boolean[width][height + 1];
		restoreWalkThroughDefaults();
	}
	
	void restoreWalkThroughDefaults()
	{
	  for (int i = 0; i < COSTS; ++i)
	  {
	    costs[i] = DEFAULT_COST;
	  } 
	  graph = null;
	}

	public int width()
	{
		return width;
	}

	public int height()
	{
		return height;
	}

	public boolean corners()
	{
		return corners;
	}

	/**
	 * Return the maximum distances from the given block to each of its neighbors.
	 * @param bx x-coordinate of the block
	 * @param by y-coordinate of the block
   * @return an array of the maximum distance to each block
	 */
	public double[] maxDists(int bx, int by)
	{
	  // Lazily compute the graph.
		if (graph == null)
		{
			graph = new SWGraphModel2(this);
		}
		return graph.maxDists(bx, by);
	}

	/**
	 * Get the shortest path to the given block from its neighbor
	 * in the given direction.
	 * @param b the block
	 * @param dir the direction to the neighbor
   * @return the vertices on the shortest path, including the start and end
   * blocks, or null if no path
	 */
	public java.util.List<Vertex> getRoute(Block b, int dir)
	{
	  return graph.getRoute(b, dir);
	}
	
	public boolean ns(int ix, int iy)
	{
		return ns[ix][iy];
	}

	public boolean ew(int ix, int iy)
	{
		return ew[ix][iy];
	}

	public void ns(int ix, int iy, boolean b)
	{
		ns[ix][iy] = b;
		graph = null;
	}

	public void ew(int ix, int iy, boolean b)
	{
		ew[ix][iy] = b;
		graph = null;
	}

	/**
	 * Return whether this intersection has been plowed
	 * from any direction.
	 * @param ix x-coordinate of the intersection
	 * @param iy y-coordinate of the intersection
	 * @return whether it has been plowed
	 */
	public boolean intersectionPlowed(int ix, int iy)
	{
		return
		  (ix != 0 && ew[ix - 1][iy]) ||
		  (ix != width && ew[ix][iy]) ||
		  (iy != 0 && ns[ix][iy - 1]) ||
		  (iy != height && ns[ix][iy]);
	}
	
	public void corners(boolean b)
	{
		corners = b;
		graph = null;
	}

	/**
	 * Utility to count the number of elements in the
	 * given 2D boolean array that are set.
	 * @param bbarray a 2D boolean array
	 * @return the count
	 */
	static int count(boolean[][] bbarray)
	{
		int ret = 0;
		for (boolean[] bb : bbarray)
		{
			for (boolean b : bb)
			{
				if (b)
				{
					++ret;
				}
			}
		}
		return ret;
	}
	
	/**
	 * Return the current score (max over all block maximum distances).
	 * @return the score
	 */
	double score()
	{
		double ret = 0;
		for (int bx = 0; bx < width; ++bx)
		{
			for (int by = 0; by < height; ++by)
			{
			  for (int dir = 0; dir < DIRS; ++dir)
			  {
			    double[] mds = maxDists(bx, by);
	        if (mds[dir] > ret)
	        {
	          ret = mds[dir];
	        }			    
			  }
			}
		}
		return ret;
	}
	
	public int totalStreetsPlowed()
	{
		return count(ns) + count(ew);
	}

	public void walkThrough(boolean b)
	{
		walkThrough = b;
		graph = null;
	}

	public boolean walkThrough()
	{
		return walkThrough;
	}
	
	public double walkThroughCost(int costType)
	{
	  return costs[costType];
	}

	public void walkThroughCost(int costType, double cost)
  {
    costs[costType] = cost;
    graph = null;
  }
}
